const _ = require('lodash')
const fs = require('fs')

const POSBITS = 8
// Offset mesa. Las mesas comienzan en 1001, por lo que se almacena la
// diferencia respecto al offset
const OFFSETMESA = 0 // 1000
const BITS_NUMERO_MESA = 14
const SMART_PACKING = true
const agrupacionesAMostrar = ['Lista'] // 'Lista' debería estar siempre

function calcBitsFijos (modo) {
  return modo + 3
}

function calcBitsVariables (modo) {
  return 7 - modo
}

function unpackSmart (binstr) {
  // Se obtiene el modo del recuento en binario y luego se lo convierte a entero
  let strModo = binstr.slice(BITS_NUMERO_MESA, BITS_NUMERO_MESA + 2)
  let modo = parseInt(strModo, 2)

  // Se calculan las cantidades de bits fijos y variables que son utilizados para
  // almacenar los distintos votos de las candidaturas
  let bitsFij = calcBitsFijos(modo)
  let bitsVar = calcBitsVariables(modo)

  // Se obtiene el largo de los datos del recuento en binario y luego se lo convierte
  // a entero
  let strCantDatos = binstr.slice(BITS_NUMERO_MESA + 2, BITS_NUMERO_MESA + 10)
  let cantDatos = parseInt(strCantDatos, 2)

  // Se obtienen los datos en binario para luego recorrerlos y obtener los distintos votos
  // de las distintas candidaturas
  let binDatos = binstr.slice(BITS_NUMERO_MESA + 10, BITS_NUMERO_MESA + 10 + cantDatos * bitsFij)

  let datos = []
  let binDato
  let i
  // Se considera el bloque de datos binario y se lo parte en distintos chunks para representar
  // los distintos votos de las candidaturas
  for (i = 0; i < binDatos.length; i += bitsFij) {
    binDato = binDatos.slice(i, i + bitsFij)
    datos.push(parseInt(binDato, 2))
  }

  // Se obtienen los datos variables (Totales y votos de las listas especiales) en binario para
  // ser procesados al final
  let binDatosVariables = binstr.slice(BITS_NUMERO_MESA + 10 + cantDatos * bitsFij)
  let longitudAjustada = Math.floor(binDatosVariables.length / (POSBITS + bitsVar)) * (POSBITS + bitsVar)
  binDatosVariables = binDatosVariables.slice(0, longitudAjustada)

  // Se considera el bloque de datos binario y se lo parte en distintos chunks para representar
  // los distintos votos de los totales y de las listas especiales
  let binDatoVariable
  let pos
  let valor
  for (i = 0; i < binDatosVariables.length; i += (POSBITS + bitsVar)) {
    binDatoVariable = binDatosVariables.slice(i, i + POSBITS + bitsVar)
    pos = parseInt(binDatoVariable.slice(0, POSBITS), 2)
    valor = parseInt(binDatoVariable.slice(POSBITS), 2) * Math.pow(2, bitsFij)
    datos[pos] = datos[pos] + valor
  }

  return datos
}

function unpackFlat (binStr, bits) {
  let datos = []
  let dato
  let i
  for (i = 0; i < binStr.length; i += bits) {
    dato = binStr.slice(i, i + bits)
    datos.push(parseInt(dato, 2))
  }
  return datos
}

function checkHex (n) {
  return /^[0-9A-Fa-f]{1,200}$/.test(n)
}

function Hex2Bin (n) {
  if (!checkHex(n)) {
    return 0
  }

  let bigString = ''
  let i
  for (i = 0; i < n.length; i += 2) {
    let part = n.slice(i, i + 2)
    let binario = parseInt(part, 16).toString(2)
    binario = (new Array(9 - binario.length)).join('0') + binario
    bigString += binario
  }
  return bigString
}

function desempaquetarQR (qrData) {
  qrData = qrData.replace(/\s/g, '')

  if (qrData.slice(0, 3) === 'REC') {
    qrData = qrData.slice(3)
  }

  let token
  let longitud
  let resultado
  if (qrData.length > 3) {
    // Se obtiene el token identificador del recuento
    token = qrData.slice(0, 2)

    // Se separa el token de los datos del recuento
    qrData = qrData.slice(2)

    // Se obtiene la longitud de los datos del recuento
    longitud = parseInt(Hex2Bin(qrData.slice(0, 4)), 2)

    // Se separa la longitud de los datos del recuento
    qrData = qrData.slice(4)
  } else {
    longitud = 999
    qrData = 0
  }

  // Si la longitud del recuento coincide con los datos, se procesa el
  // recuento.
  if (longitud === qrData.length) {
    // Se obtiene el largo de los DNIs de las Autoridades
    let lenDocAutoridades = parseInt(Hex2Bin(qrData.slice(0, 2)), 2)

    // Se obtienen los DNIs de los datos del recuento
    let docs = qrData.slice(2, lenDocAutoridades * 2 + 2)

    // Se separan los DNIs de los datos del recuento
    qrData = qrData.slice(lenDocAutoridades * 2 + 4 + 14, -16)

    let mesa
    let datos

    // Si se utiliza el SMART_PACKING se procesa de forma inteligente el
    // recuento
    if (SMART_PACKING) {
      // Paramos el hexa a binario
      let binStr = Hex2Bin(qrData)
      let strNumMesa = binStr.slice(0, BITS_NUMERO_MESA)
      mesa = parseInt(strNumMesa, 2) + OFFSETMESA
      datos = unpackSmart(binStr)
    } else {
      // Nro de Mesa (id único) de n bytes:
      let nroMesa = Math.ceil(BITS_NUMERO_MESA / 8) * 2
      var strNumMesa = qrData.slice(0, nroMesa)
      mesa = parseInt(strNumMesa, 16) + OFFSETMESA
      // extraigo la cantidad de bits de cada valor:
      let bits = parseInt(qrData.slice(nroMesa, nroMesa + 2), 16)
      qrData = qrData.slice(nroMesa + 2)
      // Pasamos el hexa a binario
      let binStr = Hex2Bin(qrData)
      datos = unpackFlat(binStr, bits)
    }
    resultado = [mesa, datos, longitud]
  } else {
    /* La longitud no matchea entonces devuelvo un error */
    resultado = [0, 'El código es incorrecto', 0]
  }

  return resultado
}

function getMesa (codigo, dirDatos) {
  let mesas = getMesas(dirDatos)
  let i

  for (i = 0; i < mesas.length; i++) {
    if (mesas[i].id_unico_mesa == codigo) {
      return mesas[i]
    }
  }
  return false
}

function getMesas (dirDatos) {
  let rawJson = fs.readFileSync(`${dirDatos}/Ubicaciones.json`)
  let jsonUbicaciones = JSON.parse(rawJson)
  let mesas = []

  let i
  let ubicacion
  for (i = 0; i < jsonUbicaciones.length; i++) {
    ubicacion = jsonUbicaciones[i]
    if (ubicacion.clase === 'Mesa') mesas.push(ubicacion)
  }

  return mesas
}

function obtenerDatosMesa (dirDatos, codDatos) {
  let rawJson
  let datosMesa = {}

  rawJson = fs.readFileSync(`${dirDatos}/${codDatos}/Agrupaciones.json`)
  datosMesa.agrupaciones = JSON.parse(rawJson)

  rawJson = fs.readFileSync(`${dirDatos}/${codDatos}/Candidaturas.json`)
  datosMesa.candidaturas = JSON.parse(rawJson)

  rawJson = fs.readFileSync(`${dirDatos}/${codDatos}/Categorias.json`)
  datosMesa.categorias = JSON.parse(rawJson)

  return datosMesa
}

function ordenarAbsolutamente (a, b) {
  let i
  for (i = 0; i < a.orden_absoluto.length; i++) {
    let valA = a.orden_absoluto[i]
    let valB = b.orden_absoluto[i]
    if (valA !== valB) {
      if (typeof (valB) === 'undefined') {
        return 1
      } else {
        return valA - valB
      }
    }
  }
  return a.orden_absoluto.length < b.orden_absoluto.length ? -1 : 1
}

function ordenarPosicion (a, b) {
  if (a.posicion < b.posicion) return -1
  else if (a.posicion > b.posicion) return 1
  else return 0
}

function buscarCandidaturaLista (candidaturas, codCategoria, codLista) {
  let i
  for (i = 0; i < candidaturas.length; i++) {
    if ((candidaturas[i].cod_categoria == codCategoria) && (candidaturas[i].cod_lista == codLista)) {
      return candidaturas[i]
    }
  }
  return false
}

function buscarCandidaturaBlanco (candidaturas, codCategoria) {
  let i
  for (i = 0; i < candidaturas.length; i++) {
    if ((candidaturas[i].cod_categoria == codCategoria) && (candidaturas[i].clase == 'Blanco')) {
      return candidaturas[i]
    }
  }
  return false
}

function buscarCandidatura (candidaturas, idCandidatura) {
  let i
  for (i = 0; i < candidaturas.length; i++) {
    if (candidaturas[i].id_candidatura == idCandidatura) {
      return candidaturas[i]
    }
  }
  return false
}

function procesarQR (qrData, dirDatos) {
  // Se desempaquetan los datos del QR
  let data = desempaquetarQR(qrData)
  let nroMesa = data[0]
  let valores = data[1]
  let longitud = data[2]

  // Se obtiene los datos de la mesa
  let mesa = getMesa(nroMesa, dirDatos)

  // Si la mesa es válida se la procesa
  if (mesa) {
    // Se obtienen los datos de la mesa como candidatos, listas, etc.
    let datosMesa = obtenerDatosMesa(dirDatos, mesa.cod_datos)

    // Se ordenan absolutamente los candidatos para poder recorrer el array de votos
    let candidatos = datosMesa.candidaturas
    candidatos = candidatos.sort(ordenarAbsolutamente)

    // Recorremos todos los candidatos y le asignamos a cada uno los votos.
    let votos = {}
    let catEspeciales = []
    let i
    let candidato
    for (i = 0; i < candidatos.length; i++) {
      candidato = candidatos[i]
      // Si es 'seleccionable' guardamos los votos.
      if (candidato.clase === 'Candidato' || candidato.clase === 'Blanco') {
        votos[candidato.id_umv] = valores.shift()
      } else if (candidato.clase === 'Especial') {
        // si es especial lo mostramos una sola vez.
        if (catEspeciales.indexOf(candidato.id_candidatura) === -1) {
          catEspeciales.push(candidato.id_candidatura)
        }
      }
    }

    // Se extraen los campos adicionales
    let cantidadSelecciones = valores.shift()
    let cantidadCiudadanos = valores.shift()

    // Se procesan los votos especiales.
    catEspeciales = _.sortBy(catEspeciales)
    let votosEspeciales = {}
    for (i = 0; i < catEspeciales.length; i++) {
      votosEspeciales[catEspeciales[i]] = valores.shift()
    }

    // Se ordenan las agrupaciones
    let agrupaciones = datosMesa.agrupaciones
    agrupaciones = agrupaciones.sort(ordenarAbsolutamente)

    // Armamos las filas
    let filas = []

    // Se genera la fila del encabezado.
    let fila = ['Clase', 'Nro.', 'Agrupación']

    // Se ordenan las categorias
    let categorias = datosMesa.categorias
    categorias = categorias.sort(ordenarPosicion)

    // Le agregamos los cargos a la fila del encabezado.
    for (i = 0; i < categorias.length; i++) {
      let categoria = categorias[i]
      fila.push(categoria.codigo)
    }
    filas.push(fila)

    // Se recorren las agrupaciones armando las filas con los votos para
    // cada categoría
    let j
    let agrupacion
    let categoria
    for (i = 0; i < agrupaciones.length; i++) {
      agrupacion = agrupaciones[i]
      fila = [
        agrupacion.clase,
        agrupacion.numero != null ? agrupacion.numero : '-',
        agrupacion.nombre
      ]
      // Se recorren las categorías para obtener los votos de los candidatos
      for (j = 0; j < categorias.length; j++) {
        candidato = null
        categoria = categorias[j]
        // Se busca el candidato para la agrupacion actual y para la categoría
        // a procesar
        if (agrupacion.clase === 'Lista') {
          candidato = buscarCandidaturaLista(
            datosMesa.candidaturas, categoria.codigo, agrupacion.codigo)

          if (!candidato) {
            fila.push('-')
          } else {
            fila.push(votos[candidato.id_umv])
          }
        }
      }
      if (agrupacionesAMostrar.indexOf(agrupacion.clase) !== -1) {
        filas.push(fila)
      }
    }

    // Se procesa la lista de votos en blanco
    fila = ['Blanco', '', 'Votos en blanco']
    // Se recorren todas las candidaturas de tipo blancas para cada
    // una de las categrías
    for (i = 0; i < categorias.length; i++) {
      categoria = categorias[i]
      candidato = buscarCandidaturaBlanco(datosMesa.candidaturas, categoria.codigo)
      if (!candidato) {
        fila.push('-')
      } else {
        fila.push(votos[candidato.id_umv])
      }
    }
    filas.push(fila)

    // Se agrega a la tabla el total de boletas contadas
    filas.push(['Total', 'Total boletas contadas', cantidadSelecciones])

    // Se procesan las listas especiales y se las agrega a la lista
    let clave
    let especial
    let ordenEspeciales = ['REC', 'IMP', 'OBS']
    for (clave of ordenEspeciales) {
      especial = buscarCandidatura(datosMesa.candidaturas, clave)
      filas.push(['Especial', especial.nombre, votosEspeciales[clave]])
    }

    // Finalmente se incluye el total de votantes
    filas.push(['Total', 'Total General', cantidadCiudadanos])

    return { filas: filas, mesa: mesa }
  } else {
    console.error(`Ocurrió un error al procesar el QR. No existe la mesa ${nroMesa}`)
  }
}

module.exports = (args, dirDatos) => {
  let QR = args._[1]
  if (!QR) {
    console.error(`El QR es inválido`)
  } else {
    let datos = procesarQR(QR, dirDatos)

    if (datos) {
      if (args.csv) {
        let pathCSV = args.csv || `mesa_${datos.mesa.numero}.csv`

        const stringify = require('csv-stringify')
        stringify(datos.filas, (err, salida) => {
          if (err) console.error(err)
          fs.writeFileSync(pathCSV, salida)
        })
      } else {
        const Table = require('cli-table')
        const tablaMesa = new Table()
        tablaMesa.push(['Mesa', `${datos.mesa.numero}`])
        tablaMesa.push(['Ciud Habilitados', `${datos.mesa.ciudadanos_habilitados}`])
        console.log(tablaMesa.toString())
        const tablaResultados = new Table({ head: datos.filas[0] })
        for (let i = 1; i < datos.filas.length; i++) tablaResultados.push(datos.filas[i])
        console.log(tablaResultados.toString())
      }
    }
  }
}
